/* OperServ functions.
 *
 * Services is copyright (c) 1996-1998 Andy Church.
 *     E-mail: <achurch@dragonfire.net>
 * This program is free but copyrighted software; see the file COPYING for
 * details.
 */

#include "services.h"

#include "os-help.c"

/*************************************************************************/

/* Nick for OperServ */
const char s_OperServ[] = OPERSERV_NAME;

/* Nick for sending global notices */
const char s_GlobalNoticer[] = GLOBAL_NOTICER_NAME;


/* Services admin list */
char *services_admins[MAX_SERVADMINS];


static int nakill = 0;
static int akill_size = 0;

static struct akill {
    char *mask;
    char *reason;
    char who[NICKMAX];
    time_t time;
    time_t expires;	/* or 0 for no expiry */
    long reserved[4];
} *akills = NULL;


struct clone {
    char *host;
    long time;
};

/* List of most recent users - statically initialized to zeros */
static struct clone clonelist[CLONE_DETECT_SIZE];

/* Which hosts have we warned about, and when?  This is used to keep us
 * from sending out notices over and over for clones from the same host. */
static struct clone warnings[CLONE_DETECT_SIZE];

/*************************************************************************/

static void add_akill(const char *mask, const char *reason, const char *who,
		      const char *expiry);
static int del_akill(const char *mask);
static void send_clone_lists(const char *source);
static void do_help(const char *source);
static void do_admin(const char *source);
static void do_akill(const char *source);
static void do_set(const char *source);
static void do_stats(const char *source);

/*************************************************************************/
/*************************************************************************/

/* Main OperServ routine. */

void operserv(const char *source, char *buf)
{
    char *cmd;
    char *s, *chan, *nick;
    int i;


    log("%s: %s: %s", s_OperServ, source, buf);
    cmd = strtok(buf, " ");


    if (!cmd) {

	return;

    } else if (stricmp(cmd, "\1PING") == 0) {

	if (!(s = strtok(NULL, "")))
	    s = "\1";
	notice(s_OperServ, source, "\1PING %s", s);

    } else if (stricmp(cmd, "MODE") == 0) {

	int argc;
	char **argv;

	s = strtok(NULL, "");
	if (!s)
	    return;
	send_cmd(s_OperServ, "MODE %s", s);
	argc = split_buf(s, &argv, 1);
	do_cmode(s_OperServ, argc, argv);

    } else if (stricmp(cmd, "KICK") == 0) {

	chan = strtok(NULL, " ");
	nick = strtok(NULL, " ");
	s = strtok(NULL, "");
	if (!chan || !nick || !s)
	    return;
	send_cmd(s_OperServ, "KICK %s %s :%s (%s)", chan, nick, source, s);

    } else if (stricmp(cmd, "HELP") == 0) {

	do_help(source);

    } else if (stricmp(cmd, "GLOBAL") == 0) {

	char *msg = strtok(NULL, "");

	if (!msg) {
	    notice(s_OperServ, source, "Syntax: \2GLOBAL \37msg\37\2");
	    notice(s_OperServ, source,
			"\2/msg %s HELP GLOBAL for more information.",
			s_OperServ);
	}
#ifdef HAVE_ALLWILD_NOTICE
	notice(s_GlobalNoticer, "$*", "%s", msg);
#else
# ifdef NETWORK_DOMAIN
	notice(s_GlobalNoticer, "$*." NETWORK_DOMAIN, "%s", msg);
# else
	/* Go through all common top-level domains.  If you have others,
	 * add them here.
	 */
	notice(s_GlobalNoticer, "$*.com", "%s", msg);
	notice(s_GlobalNoticer, "$*.net", "%s", msg);
	notice(s_GlobalNoticer, "$*.org", "%s", msg);
	notice(s_GlobalNoticer, "$*.edu", "%s", msg);
# endif
#endif
    } else if (stricmp(cmd, "ADMIN") == 0) {

	do_admin(source);

    } else if (stricmp(cmd, "AKILL") == 0) {

	do_akill(source);

    } else if (stricmp(cmd, "STATS") == 0) {

	do_stats(source);

    /* Services admins only below this point. */

    } else if (!is_services_admin(source)) {

	notice(s_OperServ, source,
		"Unrecognized command \2%s\2.  Type \"/msg %s HELP\" for help.",
		cmd, s_OperServ);

    } else if (stricmp(cmd, "LISTIGNORE") == 0) {

	int sent_header = 0;
	IgnoreData *id;

	for (i = 0; i < 256; i++) {
	    for (id = ignore[i]; id; id = id->next) {
		if (!sent_header) {
		    notice(s_OperServ, source, "Services ignorance list:");
		    sent_header = 1;
		}
		notice(s_OperServ, source, "%d %s", id->time, id->who);
	    }
	}
	if (!sent_header)
	    notice(s_OperServ, source, "Ignorance list is empty.");

    } else if (stricmp(cmd, "UPDATE") == 0) {

	notice(s_OperServ, source, "Updating databases.");
	save_data = 1;

    } else if (stricmp(cmd, "QUIT") == 0) {

	quitmsg = malloc(28 + strlen(source));
	if (!quitmsg)
	    quitmsg = "QUIT command received, but out of memory!";
	else
	    sprintf(quitmsg, "QUIT command received from %s", source);
	quitting = 1;

    } else if (stricmp(cmd, "SHUTDOWN") == 0) {

	quitmsg = malloc(32 + strlen(source));
	if (!quitmsg)
	    quitmsg = "SHUTDOWN command received, but out of memory!";
	else
	    sprintf(quitmsg, "SHUTDOWN command received from %s", source);
	save_data = 1;
	delayed_quit = 1;

    } else if (stricmp(cmd, "RESTART") == 0) {

#ifdef SERVICES_BIN
	quitmsg = malloc(31 + strlen(source));
	if (!quitmsg)
	    quitmsg = "RESTART command received, but out of memory!";
	else
	    sprintf(quitmsg, "RESTART command received from %s", source);
	raise(SIGHUP);
#else
	notice(s_OperServ, source,
		"SERVICES_BIN not defined; cannot restart.  Rerun the "
		"\2configure\2 script and recompile Services to enable "
		"the RESTART command.");
#endif

    } else if (stricmp(cmd, "RAW") == 0) {

	char *text = strtok(NULL, "");
	if (!text)
	    notice(s_OperServ, source, "Syntax: \2RAW \37command\37\2");
	else
	    send_cmd(NULL, text);

    } else if (stricmp(cmd, "JUPE") == 0) {

	char *jserver = strtok(NULL, " ");
	if (!jserver) {
	    notice(s_OperServ, source, "Syntax: \2JUPE \37servername\37\2");
	} else {
	    wallops(s_OperServ, "\2Juping\2 %s by request of \2%s\2.",
			jserver, source);
	    send_cmd(NULL, "SERVER %s 2 :Jupitered server", jserver);
	}

    } else if (stricmp(cmd, "SET") == 0) {

	do_set(source);


    /* From here on, Services root only. */

    } else if (!is_services_root(source)) {

	notice(s_OperServ, source,
		"Unrecognized command \2%s\2.  Type \"/msg %s HELP\" for help.",
		cmd, s_OperServ);

    /* Well, nothing here just yet... */

    } else {

	notice(s_OperServ, source,
		"Unrecognized command \2%s\2.  Type \"/msg %s HELP\" for help.",
		cmd, s_OperServ);

    }
}

/*************************************************************************/
/**************************** Privilege checks ***************************/
/*************************************************************************/

/* Load OperServ access lists. */

void load_os_dbase(void)
{
    FILE *f;
    int i;

    if (!(f = open_db(s_OperServ, OPERSERV_DB, "r")))
	return;
    switch (i = get_file_version(f, OPERSERV_DB)) {
      case 4:
      case 3:
	i = fgetc(f)<<8 | fgetc(f);
	while (--i >= 0)
	    services_admins[i] = read_string(f, OPERSERV_DB);
	break;
      default:
	fatal("Unsupported version (%d) on %s", i, OPERSERV_DB);
    } /* switch (version) */
    close_db(f, OPERSERV_DB);
}

/*************************************************************************/

/* Save OperServ access lists. */

void save_os_dbase(void)
{
    FILE *f;
    int i, count = 0;

    if (!(f = open_db(s_OperServ, OPERSERV_DB, "w")))
	return;
    for (i = 0; i < MAX_SERVADMINS; i++) {
	if (services_admins[i])
	    count++;
    }
    fputc(count>>8, f);
    fputc(count & 255, f);
    for (i = 0; i < MAX_SERVADMINS; i++) {
	if (services_admins[i])
	    write_string(services_admins[i], f, OPERSERV_DB);
    }
    close_db(f, OPERSERV_DB);
}

/*************************************************************************/

/* Does the given nick have Services root privileges? */

int is_services_root(const char *nick)
{
    NickInfo *ni;

    if (stricmp(nick, SERVICES_ROOT) != 0)
	return 0;
#ifdef SKELETON
    return 1;
#else
    if ((ni = findnick(nick)) && (ni->flags & NI_IDENTIFIED))
	return 1;
    return 0;
#endif
}

/*************************************************************************/

/* Does the given nick have Services admin privileges? */

int is_services_admin(const char *nick)
{
    NickInfo *ni;
    int i;

    if (is_services_root(nick))
	return 1;
    for (i = 0; i < MAX_SERVADMINS; i++) {
	if (services_admins[i] && stricmp(nick, services_admins[i]) == 0) {
#ifdef SKELETON
	    return 1;
#else
	    if ((ni = findnick(nick)) && (ni->flags & NI_IDENTIFIED))
		return 1;
	    return 0;
#endif
	}
    }
    return 0;
}

/*************************************************************************/
/****************************** AKILL stuff ******************************/
/*************************************************************************/

void load_akill(void)
{
    FILE *f;
    int i;

    if (!(f = open_db("AKILL", AKILL_DB, "r")))
	return;

    switch (i = get_file_version(f, AKILL_DB)) {

      case 4:
      case 3:
	nakill = fgetc(f) * 256 + fgetc(f);
	if (nakill < 8)
	    akill_size = 16;
	else
	    akill_size = 2*nakill;
	akills = smalloc(sizeof(*akills) * akill_size);
	if (!nakill) {
	    fclose(f);
	    return;
	}
	if (nakill != fread(akills, sizeof(*akills), nakill, f))
	    fatal_perror("Read error on %s", AKILL_DB);
	for (i = 0; i < nakill; i++) {
	    akills[i].mask = read_string(f, AKILL_DB);
	    akills[i].reason = read_string(f, AKILL_DB);
	}
	break;

      case 2: {
	struct {
	    char *mask;
	    char *reason;
	    char who[NICKMAX];
	    time_t time;
	} old_akill;
	nakill = fgetc(f) * 256 + fgetc(f);
	if (nakill < 8)
	    akill_size = 16;
	else
	    akill_size = 2*nakill;
	akills = smalloc(sizeof(*akills) * akill_size);
	if (!nakill) {
	    fclose(f);
	    return;
	}
	for (i = 0; i < nakill; i++) {
	    if (1 != fread(&old_akill, sizeof(old_akill), 1, f))
		fatal_perror("Read error on %s", AKILL_DB);
	    akills[i].time = old_akill.time;
	    strscpy(akills[i].who, old_akill.who, sizeof(akills[i].who));
	    akills[i].expires = 0;
	}
	for (i = 0; i < nakill; i++) {
	    akills[i].mask = read_string(f, AKILL_DB);
	    akills[i].reason = read_string(f, AKILL_DB);
	}
	break;
      } /* case 2 */

      case 1: {
	struct {
	    char *mask;
	    char *reason;
	    time_t time;
	} old_akill;
	nakill = fgetc(f) * 256 + fgetc(f);
	if (nakill < 8)
	    akill_size = 16;
	else
	    akill_size = 2*nakill;
	akills = smalloc(sizeof(*akills) * akill_size);
	if (!nakill) {
	    fclose(f);
	    return;
	}
	for (i = 0; i < nakill; i++) {
	    if (1 != fread(&old_akill, sizeof(old_akill), 1, f))
		fatal_perror("Read error on %s", AKILL_DB);
	    akills[i].time = old_akill.time;
	    akills[i].who[0] = 0;
	    akills[i].expires = 0;
	}
	for (i = 0; i < nakill; i++) {
	    akills[i].mask = read_string(f, AKILL_DB);
	    akills[i].reason = read_string(f, AKILL_DB);
	}
	break;
      } /* case 1 */

      default:
	fatal("Unsupported version (%d) on %s", i, AKILL_DB);
    } /* switch (version) */

    close_db(f, AKILL_DB);
}

/*************************************************************************/

void save_akill(void)
{
    FILE *f;
    int i;

    f = open_db("AKILL", AKILL_DB, "w");

    fputc(nakill/256, f);
    fputc(nakill & 255, f);
    if (fwrite(akills, sizeof(*akills), nakill, f) != nakill)
	fatal_perror("Write error on %s", AKILL_DB);
    for (i = 0; i < nakill; i++) {
	write_string(akills[i].mask, f, AKILL_DB);
	write_string(akills[i].reason, f, AKILL_DB);
    }
    close_db(f, AKILL_DB);
}

/*************************************************************************/

/* Does the user match any AKILLs? */

int check_akill(const char *nick, const char *username, const char *host)
{
    char buf[512];
    int i;
    char *host2, *username2;

    strscpy(buf, username, sizeof(buf)-2);
    i = strlen(buf);
    buf[i++] = '@';
    strlower(strscpy(buf+i, host, sizeof(buf)-i));
    for (i = 0; i < nakill; i++) {
	if (match_wild_nocase(akills[i].mask, buf)) {
#ifndef IRC_DALNET
	    time_t now = time(NULL);
#endif
	    /* Don't use kill_user(); that's for people who have already
	     * signed on.  This is called before the User structure is
	     * created.
	     */
	    send_cmd(s_OperServ,
			"KILL %s :%s (You are banned from this network)",
			nick, s_OperServ);
	    username2 = sstrdup(akills[i].mask);
	    host2 = strchr(username2, '@');
	    if (!host2) {
		/* Glurp... this oughtn't happen, but if it does, let's not
		 * play with null pointers.  Yell and bail out.
		 */
		wallops(NULL, "Missing @ in AKILL: %s", akills[i].mask);
		log("Missing @ in AKILL: %s", akills[i].mask);
		continue;
	    }
	    *host2++ = 0;
#ifdef IRC_DALNET
	    send_cmd(server_name,
		    "AKILL %s %s :You are banned from this network",
		    host2, username2);
#else
	    send_cmd(server_name,
		    "GLINE * +%d %s@%s :You are banned from this network",
		    akills[i].expires && akills[i].expires>now
				? akills[i].expires-time(NULL)
				: 999999999,
		    username2, host2);
#endif
	    free(username2);
	    return 1;
	}
    }
    return 0;
}

/*************************************************************************/

/* Delete any expired autokills. */

void expire_akills(void)
{
    int i;
    time_t now = time(NULL);
    char *s;

    for (i = 0; i < nakill; i++) {
	if (akills[i].expires == 0 || akills[i].expires > now)
	    continue;
#ifdef IRC_DALNET
	if (s = strchr(akills[i].mask, '@')) {
	    *s++ = 0;
	    strlower(s);
	    send_cmd(server_name, "RAKILL %s %s", s, akills[i].mask);
	}
#endif
	free(akills[i].mask);
	free(akills[i].reason);
	nakill--;
	if (i < nakill)
	    bcopy(akills+i+1, akills+i, sizeof(*akills) * (nakill-i));
	i--;
    }
}

/*************************************************************************/

/* Note that all parameters except expiry are assumed to be non-NULL.  A
 * value of NULL for expiry indicates that the AKILL should not expire.
 */

static void add_akill(const char *mask, const char *reason, const char *who,
		      const char *expiry)
{
    if (nakill >= akill_size) {
	if (akill_size < 8)
	    akill_size = 8;
	else
	    akill_size *= 2;
	akills = srealloc(akills, sizeof(*akills) * akill_size);
    }
    akills[nakill].mask = sstrdup(mask);
    akills[nakill].reason = sstrdup(reason);
    akills[nakill].time = time(NULL);
    strscpy(akills[nakill].who, who, NICKMAX);
    if (expiry) {
	int amount = strtol(expiry, (char **)&expiry, 10);
	if (amount == 0) {
	    akills[nakill].expires = 0;
	} else {
	    switch (*expiry) {
		case 'd': amount *= 24;
		case 'h': amount *= 60;
		case 'm': amount *= 60; break;
		default : amount = -akills[nakill].time;
	    }
	    akills[nakill].expires = amount + akills[nakill].time;
	}
    } else {
	akills[nakill].expires = AKILL_DEF_EXPIRY + akills[nakill].time;
    }
    nakill++;
}

/*************************************************************************/

/* Return whether the mask was found in the AKILL list. */

static int del_akill(const char *mask)
{
    int i;

    for (i = 0; i < nakill && strcmp(akills[i].mask, mask) != 0; i++)
	;
    if (i < nakill) {
	free(akills[i].mask);
	free(akills[i].reason);
	nakill--;
	if (i < nakill)
	    bcopy(akills+i+1, akills+i, sizeof(*akills) * (nakill-i));
	return 1;
    } else {
	return 0;
    }
}

/*************************************************************************/
/**************************** Clone detection ****************************/
/*************************************************************************/

/* We just got a new user; does it look like a clone?  If so, send out a
 * wallops.
 */

void check_clones(User *user)
{
    int i, clone_count;
    long last_time;

    if (clonelist[0].host)
	free(clonelist[0].host);
    i = CLONE_DETECT_SIZE-1;
    memmove(clonelist, clonelist+1, sizeof(struct clone) * i);
    clonelist[i].host = sstrdup(user->host);
    last_time = clonelist[i].time = time(NULL);
    clone_count = 1;
    while (--i >= 0 && clonelist[i].host) {
	if (clonelist[i].time < last_time - CLONE_MAX_DELAY)
	    break;
	if (stricmp(clonelist[i].host, user->host) == 0) {
	    ++clone_count;
	    last_time = clonelist[i].time;
	    if (clone_count >= CLONE_MIN_USERS)
		break;
	}
    }
    if (clone_count >= CLONE_MIN_USERS) {
	/* Okay, we have clones.  Check first to see if we already know
	 * about them. */
	for (i = CLONE_DETECT_SIZE-1; i >= 0 && warnings[i].host; --i) {
	    if (stricmp(warnings[i].host, user->host) == 0)
		break;
	}
	if (i < 0 || warnings[i].time < user->signon - CLONE_WARNING_DELAY) {
	    /* Send out the warning, and note it. */
	    wallops(s_OperServ,
		"\2WARNING\2 - possible clones detected from %s", user->host);
	    i = CLONE_DETECT_SIZE-1;
	    if (warnings[0].host)
		free(warnings[0].host);
	    bcopy(warnings+1, warnings, sizeof(struct clone) * i);
	    warnings[i].host = sstrdup(user->host);
	    warnings[i].time = clonelist[i].time;
	}
    }
}


/* Debug: send clone arrays to given nick. */

static void send_clone_lists(const char *source)
{
    int i;

    notice(s_OperServ, source, "clonelist[]");
    for (i = 0; i < CLONE_DETECT_SIZE; i++) {
	if (clonelist[i].host)
	    notice(s_OperServ, source, "    %10ld  %s", clonelist[i].time, clonelist[i].host ? clonelist[i].host : "(null)");
    }
    notice(s_OperServ, source, "warnings[]");
    for (i = 0; i < CLONE_DETECT_SIZE; i++) {
	if (clonelist[i].host)
	    notice(s_OperServ, source, "    %10ld  %s", warnings[i].time, warnings[i].host ? warnings[i].host : "(null)");
    }
}

/*************************************************************************/
/*********************** OperServ command functions **********************/
/*************************************************************************/

/* HELP command. */

static void do_help(const char *source)
{
    const char *cmd = strtok(NULL, " ");

    if (!cmd)
	notice_list(s_OperServ, source, os_help);
    else if (stricmp(cmd, "MODE") == 0)
	notice_list(s_OperServ, source, mode_help);
    else if (stricmp(cmd, "KICK") == 0)
	notice_list(s_OperServ, source, kick_help);
    else if (stricmp(cmd, "ADMIN") == 0)
	notice_list(s_OperServ, source, admin_help);
    else if (stricmp(cmd, "AKILL") == 0)
	notice_list(s_OperServ, source, akill_help);
    else if (stricmp(cmd, "GLOBAL") == 0) {
	/* Information varies, so we need to do it manually. */
	notice(s_OperServ, source, "Syntax: GLOBAL \37message\37");
	notice(s_OperServ, source, "");
	notice(s_OperServ, source,
			"Allows IRCops to send messages to all users on the");
	notice(s_OperServ, source,
			"network.  The message will be sent from the nick");
	notice(s_OperServ, source, "\2%s\2.", s_GlobalNoticer);
    } else if (stricmp(cmd, "STATS") == 0)
	notice_list(s_OperServ, source, stats_help);
    else if (stricmp(cmd, "SET") == 0) {
	cmd = strtok(NULL, " ");
	if (!cmd)
	    notice_list(s_OperServ, source, set_help);
	else if (stricmp(cmd, "READONLY"))
	    notice_list(s_OperServ, source, set_readonly_help);
	else
	    notice(s_OperServ, source, "Unknown SET option \2%s\2.", cmd);
    } else if (stricmp(cmd, "QUIT") == 0)
	notice_list(s_OperServ, source, quit_help);
    else if (stricmp(cmd, "UPDATE") == 0)
	notice_list(s_OperServ, source, update_help);
    else if (stricmp(cmd, "SHUTDOWN") == 0)
	notice_list(s_OperServ, source, shutdown_help);
    else if (stricmp(cmd, "RESTART") == 0)
	notice_list(s_OperServ, source, restart_help);
    else if (stricmp(cmd, "JUPE") == 0)
	notice_list(s_OperServ, source, jupe_help);
    else if (stricmp(cmd, "RAW") == 0)
	notice_list(s_OperServ, source, raw_help);
    else
	notice(s_OperServ, source,
			"No help available for command \2%s\2.", cmd);
}

/*************************************************************************/

/* Services admin list modification. */

static void do_admin(const char *source)
{
    char *cmd, *nick;
    int i;

    cmd = strtok(NULL, " ");
    if (!cmd)
	cmd = "";

    if (stricmp(cmd, "ADD") == 0) {
	if (!is_services_root(source)) {
	    notice(s_OperServ, source, "Permission denied.");
	    return;
	}
	nick = strtok(NULL, " ");
	if (nick) {
	    for (i = 0; i < MAX_SERVADMINS; i++) {
		if (!services_admins[i])
		    break;
	    }
	    if (i < MAX_SERVADMINS) {
		services_admins[i] = sstrdup(nick);
		notice(s_OperServ, source,
			"%s added to Services admin list.", nick);
	    } else {
		notice(s_OperServ, source,
			"Too many entries (%d) on Services admin list; cannot add more.",
			MAX_SERVADMINS);
	    }
	    if (readonly) {
		notice(s_OperServ, source,
		    "\2Notice:\2 Changes will not be saved!  Services is in read-only mode.");
	    }
	} else {
	    notice(s_OperServ, source, "Syntax: ADMIN ADD \37nick\37");
	}

    } else if (stricmp(cmd, "DEL") == 0) {
	if (!is_services_root(source)) {
	    notice(s_OperServ, source, "Permission denied.");
	    return;
	}
	if (nick = strtok(NULL, " ")) {
	    for (i = 0; i < MAX_SERVADMINS; i++) {
		if (services_admins[i] && stricmp(nick,services_admins[i]) == 0)
		    break;
	    }
	    if (i < MAX_SERVADMINS) {
		free(services_admins[i]);
		services_admins[i] = NULL;
		notice(s_OperServ, source,
			"%s removed from Services admin list.", nick);
		if (readonly) {
		    notice(s_OperServ, source,
			    "\2Notice:\2 Changes will not be saved!  Services is in read-only mode.");
		}
	    } else {
		notice(s_OperServ, source,
			"%s not found on Services admin list.", nick);
	    }
	} else {
	    notice(s_OperServ, source, "Syntax: ADMIN DEL \37nick\37");
	}

    } else if (stricmp(cmd, "LIST") == 0) {
	for (i = 0; i < MAX_SERVADMINS; i++) {
	    if (services_admins[i])
		notice(s_OperServ, source, "%s", services_admins[i]);
	}

    } else {
	notice(s_OperServ, source,
		"Syntax: \2ADMIN {ADD|DEL|LIST} [\37nick\37]\2");
	notice(s_OperServ, source,
		"For help: \2/msg %s HELP ADMIN\2", s_OperServ);
    }
}

/*************************************************************************/

/* AKILL list modification. */

static void do_akill(const char *source)
{
    char *cmd, *mask, *reason, *who, *expiry, *s;
    int i;

    cmd = strtok(NULL, " ");
    if (!cmd)
	cmd = "";

    if (stricmp(cmd, "ADD") == 0) {
	mask = strtok(NULL, " ");
	if (mask && *mask == '+') {
	    expiry = mask;
	    mask = strtok(NULL, " ");
	} else {
	    expiry = NULL;
	}
	if (mask && (reason = strtok(NULL, ""))) {
	    if (s = strchr(mask, '@')) {
		strlower(s);
	    } else {
		notice(s_OperServ, source, "Hostmask must contain an `@'
			character.");
		return;
	    }
	    add_akill(mask, reason, source, expiry);
	    notice(s_OperServ, source, "%s added to AKILL list.", mask);
	    if (readonly) {
		notice(s_OperServ, source,
		    "\2Notice:\2 Changes will not be saved!  Services is in read-only mode.");
	    }
	} else {
	    notice(s_OperServ, source,
			"Syntax: AKILL ADD [+\37expiry\37] \37mask\37 \37reason\37");
	}

    } else if (stricmp(cmd, "DEL") == 0) {
	if (mask = strtok(NULL, " ")) {
#ifdef IRC_DALNET
	    if (s = strchr(mask, '@'))
		strlower(s);
#endif
	    if (del_akill(mask)) {
		notice(s_OperServ, source, "%s removed from AKILL list.", mask);
#ifdef IRC_DALNET
		if (s) {
		    *s++ = 0;
		    send_cmd(server_name, "RAKILL %s %s", s, mask);
		} else {
		    /* We lose... can't figure out what's a username and what's
		     * a hostname.  Ah well.
		     */
		}
#endif
		if (readonly) {
		    notice(s_OperServ, source,
			"\2Notice:\2 Changes will not be saved!  Services is in read-only mode.");
		}
	    } else {
		notice(s_OperServ, source, "%s not found on AKILL list.", mask);
	    }
	} else {
	    notice(s_OperServ, source, "Syntax: AKILL DEL \37mask\37");
	}

    } else if (stricmp(cmd, "LIST") == 0) {
	s = strtok(NULL, " ");
	if (!s)
	    s = "*";
	if (strchr(s, '@'))
	    strlower(strchr(s, '@'));
	notice(s_OperServ, source, "Current AKILL list:");
	for (i = 0; i < nakill; i++) {
	    if (!s || match_wild(s, akills[i].mask)) {
		notice(s_OperServ, source, "%-32s  %s",
					akills[i].mask, akills[i].reason);
	    }
	}

    } else if (stricmp(cmd, "VIEW") == 0) {
	s = strtok(NULL, " ");
	if (!s)
	    s = "*";
	if (strchr(s, '@'))
	    strlower(strchr(s, '@'));
	notice(s_OperServ, source, "Current AKILL list:");
	for (i = 0; i < nakill; i++) {
	    if (!s || match_wild(s, akills[i].mask)) {
		char timebuf[32], expirebuf[64];
		struct tm tm;
		time_t t = time(NULL);

		tm = *localtime(akills[i].time ? &akills[i].time : &t);
		strftime(timebuf, sizeof(timebuf), "%d %b %Y", &tm);
		if (akills[i].expires == 0) {
		    strcpy(expirebuf, "does not expire");
		} else if (akills[i].expires <= t) {
		    strcpy(expirebuf, "expires at next database update");
		} else {
		    time_t t2 = akills[i].expires - t;
		    t2 += 59;
		    if (t2 < 3600) {
			t2 /= 60;
			snprintf(expirebuf, sizeof(expirebuf),
				"expires in %d minute%s", t2, t2==1 ? "" : "s");
		    } else if (t2 < 86400) {
			t2 /= 60;
			snprintf(expirebuf, sizeof(expirebuf),
				"expires in %d hour%s, %d minute%s",
				t2/60, t2<120 ? "" : "s",
				t2%60, t2%60==1 ? "" : "s");
		    } else {
			t2 /= 86400;
			snprintf(expirebuf, sizeof(expirebuf),
				"expires in %d day%s", t2, t2==1 ? "" : "s");
		    }
				
		}
		notice(s_OperServ, source, "%s (by %s on %s; %s)",
				akills[i].mask,
				*akills[i].who ? akills[i].who : "<unknown>",
				timebuf, expirebuf);
		notice(s_OperServ, source, "    %s", akills[i].reason);
	    }
	}

    } else {
	notice(s_OperServ, source,
		"Syntax: \2AKILL {ADD|DEL|LIST|VIEW} [[+\37expiry\37] \37mask\37 [\37reason\37]]\2");
	notice(s_OperServ, source,
		"For help: \2/msg %s HELP AKILL\2", s_OperServ);
    }
}

/*************************************************************************/

/* Set various Services runtime options. */

static void do_set(const char *source)
{
    char *option = strtok(NULL, " ");
    char *setting = strtok(NULL, " ");

    if (!option || !setting) {
	notice(s_OperServ, source,
			"Syntax: \2SET \37option\37 \37setting\37\2");
    } else if (stricmp(option, "IGNORE") == 0) {
	if (stricmp(setting, "on") == 0) {
	    allow_ignore = 1;
	    notice(s_OperServ, source, "Ignore code \2will\2 be used.");
	} else if (stricmp(setting, "off") == 0) {
	    allow_ignore = 0;
	    notice(s_OperServ, source, "Ignore code \2will not\2 be used.");
	} else {
	    notice(s_OperServ, source,
		    "Setting for \2IGNORE\2 must be \2ON\2 or \2OFF\2.");
	}
    } else if (stricmp(option, "READONLY") == 0) {
	if (stricmp(setting, "on") == 0) {
	    readonly = 1;
	    log("Read-only mode activated");
	    close_log();
	    notice(s_OperServ, source, "Services is now in read-only mode.");
	} else if (stricmp(setting, "off") == 0) {
	    readonly = 0;
	    open_log();
	    log("Read-only mode deactivated");
	    notice(s_OperServ, source, "Services is now in read-write mode.");
	} else {
	    notice(s_OperServ, source,
		    "Setting for \2READONLY\2 must be \2ON\2 or \2OFF\2.");
	}
    } else {
	notice(s_OperServ, source, "Unknown option \2%s\2.", option);
    }
}

/*************************************************************************/

/* STATS command. */

static void do_stats(const char *source)
{
    time_t uptime = time(NULL) - start_time;
    char *extra = strtok(NULL, "");

    if (extra && stricmp(extra, "ALL") != 0) {
	if (stricmp(extra, "AKILL") == 0) {
	    int timeout = AKILL_DEF_EXPIRY+59;
	    notice(s_OperServ, source,
			"Current number of AKILLs: \2%d\2", nakill);
	    if (timeout >= 86400)
		notice(s_OperServ, source,
			"Default AKILL expiry time: \2%d day%s\2",
			timeout/86400, timeout>=172800 ? "s" : "");
	    else if (timeout >= 3600)
		notice(s_OperServ, source,
			"Default AKILL expiry time: \2%d hour%s\2",
			timeout/3600, timeout>=7200 ? "s" : "");
	    else if (timeout >= 60)
		notice(s_OperServ, source,
			"Default AKILL expiry time: \2%d minute%s\2",
			timeout/60, timeout>=120 ? "s" : "");
	    else
		notice(s_OperServ, source,
			"Default AKILL expiry time: \2None\2");
	    return;
	} else {
	    notice(s_OperServ, source, "Unknown STATS option: \2%s\2", extra);
	}
    }

    notice(s_OperServ, source, "Current users: \2%d\2 (\2%d\2 ops)",
			usercnt, opcnt);
    notice(s_OperServ, source, "Maximum users: \2%d\2", maxusercnt);
    if (uptime > 86400)
	notice(s_OperServ, source,
			"Services up \2%d\2 day%s, \2%02d:%02d\2",
			uptime/86400, (uptime/86400 == 1) ? "" : "s",
			(uptime/3600) % 24, (uptime/60) % 60);
    else if (uptime > 3600)
	notice(s_OperServ, source,
			"Services up \2%d hour%s, %d minute%s\2",
			uptime/3600, uptime/3600==1 ? "" : "s",
			(uptime/60) % 60, (uptime/60)%60==1 ? "" : "s");
    else
	notice(s_OperServ, source,
			"Services up \2%d minute%s, %d second%s\2",
			uptime/60, uptime/60==1 ? "" : "s",
			uptime%60, uptime%60==1 ? "" : "s");

    if (extra && stricmp(extra, "ALL") == 0 && is_services_admin(source)) {
	long count, mem;
	int i;

	get_user_stats(&count, &mem);
	notice(s_OperServ, source,
			"User    : \2%6d\2 records, \2%5d\2 kB",
			count, (mem+512) / 1024);
	get_channel_stats(&count, &mem);
	notice(s_OperServ, source,
			"Channel : \2%6d\2 records, \2%5d\2 kB",
			count, (mem+512) / 1024);
	get_nickserv_stats(&count, &mem);
	notice(s_OperServ, source,
			"NickServ: \2%6d\2 records, \2%5d\2 kB",
			count, (mem+512) / 1024);
	get_chanserv_stats(&count, &mem);
	notice(s_OperServ, source,
			"ChanServ: \2%6d\2 records, \2%5d\2 kB",
			count, (mem+512) / 1024);
	get_memoserv_stats(&count, &mem);
	notice(s_OperServ, source,
			"MemoServ: \2%6d\2 records, \2%5d\2 kB",
			count, (mem+512) / 1024);
	count = 0;
	mem = sizeof(struct clone) * CLONE_DETECT_SIZE * 2;
	for (i = 0; i < CLONE_DETECT_SIZE; i++) {
	    if (clonelist[i].host) {
		count++;
		mem += strlen(clonelist[i].host)+1;
	    }
	    if (warnings[i].host) {
		count++;
		mem += strlen(warnings[i].host)+1;
	    }
	}
	for (i = 0; i < MAX_SERVADMINS; i++) {
	    if (services_admins[i]) {
		count++;
		mem += strlen(services_admins[i])+1;
	    }
	}
	mem += sizeof(struct akill) * akill_size;
	count += nakill;
	for (i = 0; i < nakill; i++)
	    mem += strlen(akills[i].mask)+1 + strlen(akills[i].reason)+1;
	notice(s_OperServ, source,
			"OperServ: \2%6d\2 records, \2%5d\2 kB",
			count, (mem+512) / 1024);
    }
}

/*************************************************************************/
